'''
EN.540.635 Software Carpentry
Lab 8 - Conjugate Gradient Solutions

Note: f2 and g2 have been changed from the original assignment - the
original functions will showcase divergent behavior in the conjugate gradient
method.
'''
import numpy as np
from scipy.optimize import minimize
from matplotlib import pyplot as plt


def f1(x1, x2):
    '''
    Some arbitrary quadratic function.

    **Parameters**
        x1: *float*
        x2: *float*

    **Returns**
        y: *float*
    '''
    return (x1 - 2.0) ** 2 + (x2 - 3.0) ** 2


def g1(x1, x2):
    '''
    The gradient of f1.

    **Parameters**
        x1: *float*
        x2: *float*

    **Returns**
        y: *list, float*
    '''
    return [2.0 * (x1 - 2.0), 2.0 * (x2 - 3.0)]


def f2(x1):
    '''
    Some arbitrary trig function.

    **Parameters**

        x1: *float*

    **Returns**

        y: *float*
    '''
    return np.sin(x1)


def g2(x1):
    '''
    The gradient of f2.

    **Parameters**
        x1: *float*

    **Returns**
        y: *float*
    '''
    return np.cos(x1)


def ensure_array_like(value):
    '''
    This function simply ensures that an input is an array.

    **Parameters**

        value: *...*
            Some object that should be array like.

    **Returns**

        value: *list or array like*
            The value now has a length!
    '''
    if not hasattr(value, "__len__"):
        value = [value]
    return value


def steepest_descent(func, grad, x0, MAXITER=1000000,
                     step_size=0.001, eps=1E-6):
    '''
    The steepest descent function, used to minimize some function f with
    its associated gradient g.

    **Parameters**

        func: *function*
            This is the function we want to minimize.
        grad: *function*
            This is the gradient of func.
        x0: *list, float*
            This is the starting coordinate for our optimizer.
        MAXITER: *int, optional*
            The maximum iteration our optimizer may run for.
        step_size: *float, optional*
            The scalar applied to the steepest descent step.
        eps: *float, optional*
            The tolerance for the gradient to end the Steepest Descent.

    **Returns**

        xopt: *list, float*
            The x at the minimum of func.
        fopt: *float*
            The function evaluation at xopt.
    '''
    # Handle if we input a value instead of an array
    # Handle if we input a value instead of an array
    x0 = np.array(ensure_array_like(x0))
    # Initialize flags
    GRAD_CONVERGED = False
    MAXITER_REACHED = False
    # Initialize x variable
    x = x0[:]
    for i in range(MAXITER):
        # Calculate gradient
        gradient = np.array(ensure_array_like(grad(*x)))
        # Check if gradient has converged to 0
        GRAD_CONVERGED = all([abs(g) < eps for g in gradient])
        if GRAD_CONVERGED:
            break
        # Calculate step
        step = -1.0 * step_size * gradient
        # Update x variable
        x += step
    # Define boolean for if we reached the maximum number of iterations
    MAXITER_REACHED = i == MAXITER - 1
    # Benchmarking print statements
    if GRAD_CONVERGED:
        print("Gradient converged below %s!" % str(eps))
        print("Iterations to Convergence: %d" % i)
    elif MAXITER_REACHED:
        print("Function has not converged yet, maxiter was reached.")
    else:
        raise Exception("Error with flags.")
    return x, func(*x)


def conjugate_gradient(func, grad, x0, MAXITER=1000000,
                       step_size=0.001, eps=1E-6):
    '''
    The conjugate gradient function, used to minimize some function f with its
    associated gradient g and a parameter beta, which is defined using the
    Fletcher-Reeves method.

        **Parameters**

        func: *function*
            This is the function we want to minimize.
        grad: *function*
            This is the gradient of func.
        x0: *list, float*
            This is the starting coordinate for our optimizer.
        MAXITER: *int, optional*
            The maximum iteration our optimizer may run for.
        step_size: *float, optional*
            The scalar applied to the steepest descent step.
        eps: *float, optional*
            The tolerance for the gradient to end the Conjugate Gradient.

    **Returns**

        xopt: *list, float*
            The x at the minimum of func.
        fopt: *float*
            The function evaluation at xopt.
    '''
    # Handle if we input a value instead of an array
    x0 = np.array(ensure_array_like(x0))
    # Define and initialize flags
    GRAD_CONVERGED = False
    MAXITER_REACHED = False
    x = x0[:]
    # Run minimization within the defined maximum number of iterations
    for i in range(MAXITER):
        # Calculate the gradient
        gradient = np.array(ensure_array_like(grad(*x)))
        # Check for convergence
        GRAD_CONVERGED = all([abs(g) < eps for g in gradient])
        if GRAD_CONVERGED:
            break
        # For the first iteration, calculate the step to take and store the
        # values of the step and the gradient for use in the following
        # iterations
        if i == 0:
            step = -1.0 * step_size * gradient
            prev_step = step
            prev_gradient = gradient
        # For the following iterations, calculate the beta parameter and use
        # it with the previous step value to adjust the new step
        else:
            beta = float(np.dot(gradient, gradient) / np.dot(
                prev_gradient, prev_gradient))
            step = [-1.0 * g + beta * h for g, h in zip(
                gradient, prev_step)]
            # Store the step and gradient values for use in the next step
            prev_step = step
            prev_gradient = gradient
        # Calculate the new values of x based on the step size
        x += step
    MAXITER_REACHED = i == MAXITER - 1
    if GRAD_CONVERGED:
        print("Gradient converged below %s!" % str(eps))
        print("Iterations to convergence: %d" % i)
    elif MAXITER_REACHED:
        print("Function has not converged yet. Maxiter was reached.")
    else:
        print("This... should not print.")
    return x, func(*x)


def calculate_R2(xvals, y_obs, m, b):
    '''
    The R2 value for some given sampled points and slope/y-intercept.

    **Parameters**

        xvals: *list, float*
            A list of x values we sampled.
        y_obs: *list, float*
            A list of the sampled y values corresponding to the given
            x values.
        m: *float*
            A slope
        b: *float*
            A y-intercept.

    **Returns**

        R2: *float*
            The R2 value.

    **References**

        - https://en.wikipedia.org/wiki/Coefficient_of_determination
    '''
    y_mean = sum(y_obs) / len(y_obs)
    total_sum_of_errors = sum([
        (y - y_mean) ** 2 for y in y_obs
    ])
    residual_sum_of_errors = sum([
        (y - (m * x + b)) ** 2 for x, y in zip(xvals, y_obs)
    ])
    return 1.0 - residual_sum_of_errors / total_sum_of_errors


def err_func(x, *args):
    '''
    Error function to be minimized.

    **Parameters**

        x: *list, float*
            The parameters to be parameterized.
        args: *list, float*
            Fixed parameters need to specify the function.

    **Returns**

        result: *float*
            The output of the error function
    '''
    # Calculate the y values from a certain linear function
    y_model = np.array(list(map(lambda r: x[0] * r + x[1], *args)))
    return np.sqrt(np.mean(np.square(y_obs - y_model)))


if __name__ == "__main__":
    ##########################################################################
    # STEP 1 - Implement your own Conjugate Gradient
    ##########################################################################
    print("\nIMPLEMENTING CONJUGATE GRADIENT")
    x0 = [200.5, 13.5]
    print("\nTesting Steepest Descent for f1 with %s" % str(x0))
    xopt, fopt = steepest_descent(f1, g1, x0)
    print("\nTesting Conjugate Gradient for f1 with %s" % str(x0))
    xopt, fopt = conjugate_gradient(f1, g1, x0)
    print("\noptimal x = " + str(xopt) + " where f1 = " + str(fopt))
    x0 = 0.1
    print("\nTesting Steepest Descent for f2 with %s" % str(x0))
    xopt, fopt = steepest_descent(f2, g2, x0)
    print("\nTesting Conjugate Gradient for f2 with %s" % str(x0))
    xopt, fopt = conjugate_gradient(f2, g2, x0)
    print("\noptimal x = " + str(xopt) + " where f2 = " + str(fopt))
    print("\n----------------------------------------------------------------------------------------")
    ##########################################################################
    # STEP 2 - Implement Scipy
    ##########################################################################
    print("IMPLEMENTING SCIPY")
    x0 = [200.5, 13.5]
    print("\nTesting Scipy optimization for f1 with %s" % str(x0))
    optres = minimize(lambda x: f1(*x), np.array(x0))
    print("Result:\n" + str(optres))
    x0 = 0.3
    print("\nTesting Scipy optimization for f2 with %s" % str(x0))
    optres = minimize(f2, np.array(x0))
    print("Result:\n" + str(optres))
    print("\n----------------------------------------------------------------------------------------")
    ##########################################################################
    # STEP 3 - Fit the line
    ##########################################################################
    print("FITTING A CURVE")
    # Given data
    x_vals = [0.00, 0.50, 1.00, 1.50, 2.00, 2.50, 3.00, 3.50, 4.00,
              4.50, 5.00, 5.50, 6.00, 6.50, 7.00, 7.50, 8.00, 8.50,
              9.00, 9.50]
    y_obs = [1.39, 0.25, -0.65, -2.38, -2.59, -5.03, -5.72, -6.28,
             -8.32, -9.49, -9.83, -12.01, -12.32, -13.21, -14.50,
             -15.51, -17.80, -18.41, -19.87, -21.31]
    # Minimze the error function to calculate parameters
    optres = minimize(err_func, (-1.0, 1.0), args=(x_vals))
    m, b = optres.x
    print("Result:\n" + str(optres))
    # Calculate model values
    x_range = np.linspace(min(x_vals), max(x_vals))
    y_model = [m * i + b for i in x_range]
    # Calculate R2 value
    R2 = calculate_R2(x_vals, y_obs, m, b)
    # Plotting
    label = ("y = %.2fx + %.2f, R2 = %.3f" % (m, b, R2))
    plt.figure()
    plt.plot(x_vals, y_obs, 'k.', label="Observed Data")
    plt.plot(x_range, y_model, 'b-', label=label)
    plt.title("Best fit of sampled data using Scipy")
    plt.xlabel("X Values")
    plt.ylabel("Y Values")
    plt.legend()
    plt.show()
